@@ -11,7 +11,8 @@ use ckb_types::{
     bytes::Bytes,
     core::{
         capacity_bytes, cell::ResolvedTransaction, BlockBuilder, BlockNumber, BlockView, Capacity,
-        EpochExt, HeaderBuilder, HeaderView, TransactionBuilder, TransactionView, UncleBlockView,
+        EpochExt, EpochNumberWithFraction, HeaderBuilder, HeaderView, TransactionBuilder,
+        TransactionView, UncleBlockView,
     },
     packed::{Byte32, CellDep, CellInput, CellOutputBuilder, OutPoint, ProposalShortId, Script},
     prelude::*,
@@ -159,7 +160,8 @@ pub fn test_should_have_no_output_in_cellbase_no_finalization_target() {
 
 #[test]
 fn test_epoch_number() {
-    let block = BlockBuilder::default().epoch(2u64.pack()).build();
+    let actual_epoch = EpochNumberWithFraction::new(2, 0, 1);
+    let block = BlockBuilder::default().epoch(actual_epoch.pack()).build();
     let mut epoch = EpochExt::default();
     epoch.set_length(1);
 
